/* logs.h -- Major data structures and core log rotation functions. */
/* $Id: logs.h 145 2002-08-05 00:06:00Z eagle $ */

#ifndef H_LOGS
#define H_LOGS

#include "config.h"
#include <sys/types.h>

/* This global variable says whether to use bzip2 instead of gzip for log
   compression. */
extern bool use_bzip2;

/* A linked list of archive locations and the number of old logs to save (if
   any) for a log. */
struct archive {
    char *path;                 /* Path to the archive location. */
    char *template;             /* Pre-expanded path to the archive. */
    int count;                  /* Number of old logs to save. */

    struct archive *next;
};

/* A linked list of logs to rotate, along with the directories to move them
   into and the mode and ownership to give the new log. */
struct log {
    char *name;                 /* Name of the local logfile. */
    char *path;                 /* Path to the local logfile. */
    char *tmppath;              /* Temporary local file for rotation. */
    struct archive *archive;    /* Locations to archive a copy of it. */
    struct command *analyze;    /* Command to run on the raw log. */
    struct command *filter;     /* Command to filter the log through. */
    bool rotate;                /* Whether to rotate the log. */
    uid_t owner;                /* Owner of the new logfile. */
    gid_t group;                /* Group of the new logfile. */
    mode_t mode;                /* Permissions for the new logfile. */

    struct log *next;
};

/* A command is an argv vector (a NULL-terminated array of arguments for the
   command to run), a count of the length of the array, and the amount of
   allocated memory for it.  We support a linked list of commands for analyze
   commands. */
struct command {
    int argc;                   /* Number of arguments. */
    int size;                   /* Allocated number of arguments. */
    char **argv;                /* NULL-terminated array of arguments. */

    struct command *next;
};

/* A process (either a "run" or a "hup" command, depending on which fields
   are filled in).  It's a "hup" command if there's a pidfile and a "run"
   command if arguments is filled in. */
struct process {
    char *pidfile;              /* PID file for a hup command. */
    struct command *command;    /* An external command to run. */
};

/* A linked list of classes of logs.  Each class stores the stop and restart
   commands for the server and a pointer to a list of logs to rotate. */
struct class {
    char *name;                 /* Name of the log class. */
    struct process *stop;       /* How to stop the server. */
    struct process *restart;    /* How to restart the server. */
    struct log *logs;           /* List of logs to rotate. */

    struct class *next;
};

/* Read in a configuration file and return a list of log classes. */
extern struct class *read_config(const char *filename);

/* Perform all of the archiving desired for a given log. */
extern bool archive_logs(struct log *log);

/* Perform archiving for a log that doesn't need to be rotated. */
extern void copy_log(const struct log *log);

/* Run a program on a log file. */
extern void analyze_log(const struct log *log);

#endif /* H_LOGS */
