/* main.c -- Log file rotation and periodic maintenance.
   $Id: main.c 207 2005-10-12 22:46:44Z rra $

   Written by Russ Allbery <rra@stanford.edu>
   Copyright 1997, 1998, 1999, 2000, 2002, 2004
       Board of Trustees, Leland Stanford Jr. University

   This program is free software.  See README included in the distribution
   for terms and conditions.

   This program is designed to rotate, truncate, or otherwise deal with
   system logs on a regular basis, including correctly stopping and
   restarting daemons so that the logs can be moved and compressed safely.
   It takes a configuration file specifying which logs to rotate and what to
   do with them. */

#include "system.h"
#include "file.h"
#include "logs.h"
#include "process.h"
#include "utility.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

/* Global variable saying whether to use bzip2 for log compression. */
bool use_bzip2 = false;


/* From a process struct, either send HUP to a process or run a command. */
static void
process(struct process *action)
{
    if (action->pidfile)
        hup(action->pidfile);
    else if (action->command)
        run(action->command->argv, NULL);
}


/* First, read in the configuration file, either the hard-coded value or the
   first command-line argument.  This returns a linked list of log classes.
   Next, step through each log class, calling the stop command if necessary,
   rotate each of the logs in that class, call the restart command, and for
   each log copy it to its archive locations while compressing.

   See logs.h for our data structure here.  We don't bother freeing memory
   as we go; there's no point, since we build up this structure, execute it,
   and then exit. */
int
main(int argc, char **argv)
{
    struct class *class, *current;
    struct log *log;
    int i;
    const char *conffile;
    bool check_only = false;
    struct stat st;

    /* Look for command-line options. */
    for (i = 1; i < argc; i++) {
        if (argv[i][0] != '-')
            break;
        if (argv[i][1] == 'h' || !strcmp(argv[i], "--help")) {
            usage(0);
        } else if (argv[i][1] == 'v' || !strcmp(argv[i], "--version")) {
            printf("%s\n", PACKAGE_STRING);
            exit(0);
        } else if (argv[i][1] == 'c' || !strcmp(argv[i], "--check-config")) {
            check_only = true;
        } else if (argv[i][1] == 'b' || !strcmp(argv[i], "--bzip2")) {
            use_bzip2 = true;
        } else {
            usage(1);
        }
    }
    if (argc > 1 && argc - i > 1)
        usage(1);

    /* If we have a command-line argument, assume that's the config file. */
    conffile = (argc - i > 0) ? argv[i] : CONFIG_FILE;
    class = read_config(conffile);
    if (check_only) {
        printf("Syntax of %s is valid\n", conffile);
        exit(0);
    }

    for (current = class; current; current = current->next) {
        if (current->stop)
            process(current->stop);

        /* Rotate all the logs that have archive directives.  Skip anything
           that isn't a regular file; usually this means that a directory was
           mistakenly specified instead of the log file, and we can't recreate
           non-regular files properly anyway. */
        for (log = current->logs; log; log = log->next) {
            if (!log->rotate)
                continue;
            if (stat(log->path, &st) < 0)
                syswarn("Cannot stat %s", log->path);
            else if (!S_ISREG(st.st_mode)) {
                warn("%s is not a regular file", log->path);
                continue;
            }
            if (log->archive->template == NULL)
                truncate_file(log->path, log->owner, log->group, log->mode);
            else {
                log->tmppath = rename_unique(log->path);
                if (log->tmppath)
                    create_file(log->path, log->owner, log->group, log->mode);
            }
        }

        if (current->restart)
            process(current->restart);

        for (log = current->logs; log; log = log->next) {
            if (log->tmppath == NULL && log->rotate) {
                if (log->archive->template != NULL)
                    warn("Skipping archiving of %s", log->name);
            } else {
                if (log->analyze != NULL)
                    analyze_log(log);
                if (log->rotate)
                    archive_logs(log);
                else if (log->archive != NULL)
                    copy_log(log);
            }
        }
    }

    /* Success.  Exit. */
    exit(0);
}
